import requests

from pytest_httpserver import HTTPServer


def test_querystring_str(httpserver: HTTPServer):
    httpserver.expect_request("/foobar", query_string="foo=bar", method="GET").respond_with_data("example_response")
    response = requests.get(httpserver.url_for("/foobar?foo=bar"))
    httpserver.check_assertions()
    assert response.text == "example_response"
    assert response.status_code == 200


def test_querystring_bytes(httpserver: HTTPServer):
    httpserver.expect_request("/foobar", query_string=b"foo=bar", method="GET").respond_with_data("example_response")
    response = requests.get(httpserver.url_for("/foobar?foo=bar"))
    httpserver.check_assertions()
    assert response.text == "example_response"
    assert response.status_code == 200


def test_querystring_dict(httpserver: HTTPServer):
    httpserver.expect_request("/foobar", query_string={"k1": "v1", "k2": "v2"}, method="GET").respond_with_data(
        "example_response"
    )
    response = requests.get(httpserver.url_for("/foobar?k1=v1&k2=v2"))
    httpserver.check_assertions()
    assert response.text == "example_response"
    assert response.status_code == 200

    response = requests.get(httpserver.url_for("/foobar?k2=v2&k1=v1"))
    httpserver.check_assertions()
    assert response.text == "example_response"
    assert response.status_code == 200



from contextlib import contextmanager
from copy import deepcopy
from multiprocessing.pool import ThreadPool
from urllib.parse import urlparse

import pytest
import requests

from pytest_httpserver import BlockingHTTPServer


@contextmanager
def when_a_request_is_being_sent_to_the_server(request):
    with ThreadPool(1) as pool:
        yield pool.apply_async(requests.request, kwds=request)


def then_the_server_gets_the_request(server, request):
    request = deepcopy(request)
    replace_url_with_uri(request)

    return server.assert_request(**request)


def replace_url_with_uri(request):
    request["uri"] = get_uri(request["url"])
    del request["url"]


def get_uri(url):
    url = urlparse(url)
    return "?".join(item for item in [url.path, url.query] if item)


def when_the_server_responds_to(client_connection, response):
    client_connection.respond_with_json(response)


def then_the_response_is_got_from(server_connection, response):
    assert server_connection.get(timeout=9).json() == response


@pytest.fixture
def httpserver():
    server = BlockingHTTPServer(timeout=1)
    server.start()

    yield server

    server.clear()
    if server.is_running():
        server.stop()


def test_behave_workflow(httpserver: BlockingHTTPServer):
    request = dict(
        method="GET",
        url=httpserver.url_for("/my/path"),
    )

    with when_a_request_is_being_sent_to_the_server(request) as server_connection:

        client_connection = then_the_server_gets_the_request(httpserver, request)

        response = {"foo": "bar"}

        when_the_server_responds_to(client_connection, response)

        then_the_response_is_got_from(server_connection, response)


def test_raises_assertion_error_when_request_does_not_match(httpserver: BlockingHTTPServer):
    request = dict(
        method="GET",
        url=httpserver.url_for("/my/path"),
    )

    with when_a_request_is_being_sent_to_the_server(request):

        with pytest.raises(AssertionError) as exc:
            httpserver.assert_request(uri="/not/my/path/")

        assert "/not/my/path/" in str(exc)
        assert "does not match" in str(exc)


def test_raises_assertion_error_when_request_was_not_sent(httpserver: BlockingHTTPServer):
    with pytest.raises(AssertionError) as exc:
        httpserver.assert_request(uri="/my/path/", timeout=1)

    assert "/my/path/" in str(exc)
    assert "timed out" in str(exc)


def test_ignores_when_request_is_not_asserted(httpserver: BlockingHTTPServer):
    request = dict(
        method="GET",
        url=httpserver.url_for("/my/path"),
    )

    with when_a_request_is_being_sent_to_the_server(request) as server_connection:

        assert server_connection.get(timeout=9).text == "No handler found for this request"


def test_raises_assertion_error_when_request_was_not_responded(httpserver: BlockingHTTPServer):
    request = dict(
        method="GET",
        url=httpserver.url_for("/my/path"),
    )

    with when_a_request_is_being_sent_to_the_server(request):

        then_the_server_gets_the_request(httpserver, request)

        httpserver.stop()  # waiting for timeout of waiting for the response

        with pytest.raises(AssertionError) as exc:
            httpserver.check_assertions()

        assert "/my/path" in str(exc)
        assert "no response" in str(exc).lower()



import requests

from pytest_httpserver import HTTPServer


def test_ordered_ok(httpserver: HTTPServer):
    httpserver.expect_ordered_request("/foobar").respond_with_data("OK foobar")
    httpserver.expect_ordered_request("/foobaz").respond_with_data("OK foobaz")

    assert len(httpserver.ordered_handlers) == 2

    # first requests should pass
    response = requests.get(httpserver.url_for("/foobar"))
    httpserver.check_assertions()
    assert response.status_code == 200
    assert response.text == "OK foobar"

    response = requests.get(httpserver.url_for("/foobaz"))
    httpserver.check_assertions()
    assert response.status_code == 200
    assert response.text == "OK foobaz"

    assert len(httpserver.ordered_handlers) == 0

    # second requests should fail due to 'oneshot' type
    assert requests.get(httpserver.url_for("/foobar")).status_code == 500
    assert requests.get(httpserver.url_for("/foobaz")).status_code == 500


def test_ordered_invalid_order(httpserver: HTTPServer):
    httpserver.expect_ordered_request("/foobar").respond_with_data("OK foobar")
    httpserver.expect_ordered_request("/foobaz").respond_with_data("OK foobaz")

    assert len(httpserver.ordered_handlers) == 2

    # these would not pass as the order is different
    # this mark the whole thing 'permanently failed' so no further requests must pass
    response = requests.get(httpserver.url_for("/foobaz"))
    assert response.status_code == 500

    response = requests.get(httpserver.url_for("/foobar"))
    assert response.status_code == 500

    # as no ordered handlers are triggered yet, these must be intact..
    assert len(httpserver.ordered_handlers) == 2
